from collections import namedtuple
import copy

class bcolors:
    RED     = '\033[41m'
    BLUE    = '\033[44m'
    GREEN   = '\033[42m'
    CYAN    = '\033[106m'
    MAGENTA = '\033[45m'
    BLACK   = '\033[40m'
    YELLOW  = '\033[103m'
    LGREY   = '\033[47m'
    ENDC    = '\033[0m'

def extract(value, s, e):
    return extract_no_shift(value, s, e) >> s

def extract_no_shift(value, s, e):
    mask = ((1<<(e + 1))-1) & ~((1<<s) - 1)
    return (value & mask)

def split_range_into_int_values(memory, value_size):
    values = []
    for u in range(0, len(memory), value_size):
        values.append(int.from_bytes(memory[u:u+value_size], 'little'))
    return values

def read_arbitrary_page(machine, addr, page_size):
    return machine.read_physical_memory(addr, page_size)

def read_page(machine, addr):
    return read_arbitrary_page(machine, addr, 4096)

def read_16k_page(machine, addr):
    return read_arbitrary_page(machine, addr, 16 * 1024)

def read_64k_page(machine, addr):
    return read_arbitrary_page(machine, addr, 64 * 1024)

def make_canonical(va, top_bit_pos = 48):
    shift = top_bit_pos - 1
    bit = (va >> shift) & 0x1
    mask = ((((2**64)-1) >> shift) * bit) << shift
    return va | mask

PagePrintSettings = namedtuple('PagePrintSettings', ['va_len', 'page_size_len', 'phys_len'])
PrintConfig = PagePrintSettings(va_len = 18, page_size_len = 14, phys_len = 12)

class Page():
    def __init__(self):
        self.va = None
        self.page_size = None
        self.w = None
        self.x = None
        self.s = None
        self.wb = None
        self.uc = None
        self.phys = None
        self.sizes = None

    def cut_after(self, cut_addr):
        i = 0
        off = 0
        while i < len(self.phys):
            if cut_addr < self.va + off + self.sizes[i]:
                break
            off += self.sizes[i]
            i += 1
        if i > 0:
            self.phys = self.phys[i:]
            self.sizes = self.sizes[i:]
        delta = 0
        if len(self.phys) >= 1 and cut_addr >= self.va:
            delta = cut_addr - (self.va + off)
            self.sizes[0] = self.sizes[0] - delta
            self.phys[0] = self.phys[0] + delta
        self.page_size = self.page_size - delta - off
        self.va = max(self.va, cut_addr)

    def cut_before(self, cut_addr):
        i = len(self.phys) - 1
        off = 0
        while i >= 0:
            if self.va < cut_addr:
                break
            off += self.sizes[i]
            i -= 1
        if i > 0:
            self.phys = self.phys[:i]
            self.sizes = self.sizes[:i]
        delta = 0
        if len(self.phys) >= 1:
            delta = max(0, (self.va + self.page_size - off) - cut_addr)
            self.sizes[-1] = self.sizes[-1] - delta
        self.page_size = min(self.page_size, cut_addr - self.va)

    def read_memory(self, machine):
        memory = b""
        for phys_range_start, phys_range_size in zip(self.phys, self.sizes):
            memory += machine.read_physical_memory(phys_range_start, phys_range_size)
        return memory

    def pwndbg_is_writeable(self):
        return self.w

    def pwndbg_is_executable(self):
        return self.x

    def to_string(self, phys_verbose):
        prefix = ""
        if not self.s:
            prefix = bcolors.CYAN + " " + bcolors.ENDC
        elif self.s:
            prefix = bcolors.MAGENTA + " " + bcolors.ENDC

        varying_str = None
        if phys_verbose:
            fmt = f"{{:>{PrintConfig.va_len}}} : {{:>{PrintConfig.page_size_len}}} : {{:>{PrintConfig.phys_len}}}"
            varying_str = fmt.format(hex(self.va), hex(self.page_size), hex(self.phys[0]))
        else:
            fmt = f"{{:>{PrintConfig.va_len}}} : {{:>{PrintConfig.page_size_len}}}"
            varying_str = fmt.format(hex(self.va), hex(self.page_size))

        s = f"{varying_str} | W:{int(self.w)} X:{int(self.x)} S:{int(self.s)} UC:{int(self.uc)} WB:{int(self.wb)}"

        res = ""
        if self.x and self.w:
            res = prefix + bcolors.BLUE + " " + s + bcolors.ENDC
        elif self.w and not self.x:
            res = prefix + bcolors.GREEN + " " + s + bcolors.ENDC
        elif self.x:
            res = prefix + bcolors.RED + " " + s + bcolors.ENDC
        else:
            res = prefix + " " + s

        return res

def merge_cont_pages(pages, func_semantic_sim, require_physical_contiguity):
    if len(pages) <= 1:
        return pages

    # Here I am just going to abuse the Page structure to contain the range
    merged_pages = []
    cur_page = copy.copy(pages[0])
    for page in pages[1:]:

        merge_pages = True
        if not (cur_page.va + cur_page.page_size == page.va and func_semantic_sim(cur_page, page)):
            merge_pages = False

        if require_physical_contiguity and not (cur_page.phys[-1] + cur_page.sizes[-1] == page.phys[0]):
            merge_pages = False

        if merge_pages:
            cur_page.page_size = cur_page.page_size + page.page_size
            if cur_page.phys[-1] + cur_page.sizes[-1] == page.phys[0]:
                # Depending on the flag require_physical_contiguity, the extended ranges may or may not be physically contiguous
                assert(len(page.phys) == 1)
                assert(len(page.sizes) == 1)
                cur_page.sizes[-1] = cur_page.sizes[-1] + page.page_size
            else:
                # If not, then add a new entry
                cur_page.phys.extend(page.phys)
                cur_page.sizes.extend(page.sizes)
        else:
            merged_pages.append(cur_page)
            cur_page = copy.copy(page)
    merged_pages.append(cur_page)
    return merged_pages 

def optimize(gig_pages, mb_pages, kb_pages, func_semantic_sim, require_physical_contiguity):
    pages = sorted(gig_pages + mb_pages + kb_pages, key = lambda p: p.va)
    opt = merge_cont_pages(pages, func_semantic_sim, require_physical_contiguity)
    return opt

def select_color(w, x, r):
    if x and w:
        return bcolors.BLUE
    if x:
        return bcolors.RED
    if w:
        return bcolors.GREEN
    if r:
        return bcolors.LGREY
    return bcolors.BLACK

def create_compound_filter(filters):
    def apply_filters(p):
        res = True
        for func in filters:
            res = res and func(p)
        return res
    return apply_filters

def search_memory(machine, page_ranges, to_search, to_search_num, aligned_to, aligned_offset):
    done_searching = False
    for range in page_ranges:
        if done_searching:
            break

        data = None
        try:
            data = range.read_memory(machine)
        except OSError:
            pass

        if data is not None:
            idx = 0
            while True:
                idx = data.find(to_search, idx)
                if idx != -1:
                    if (idx - aligned_offset) % aligned_to == 0:
                        yield (range.va + idx, range)
                        to_search_num = to_search_num - 1
                        if to_search_num == 0:
                            done_searching = True
                            break
                    idx = idx + 1
                else:
                    break
    return None

def find_aliases(virtual_page_ranges, phys_verobse):
    # First collect the physical ranges, aka inverse virtual map
    phys_ranges = []
    i = 0
    for range in virtual_page_ranges:
        virtual_page_range_base = range.va
        off = 0
        for phys_range, phys_range_size in zip(range.phys, range.sizes):
            phys_ranges.append((phys_range, phys_range + phys_range_size, virtual_page_range_base + off))
            off = off + phys_range_size

    # Sort the physical ranges
    phys_ranges = sorted(phys_ranges, key=lambda key: key[0])

    # TODO
    # We could use bisect here to speed-up
    # The first loop can be simplified
    # The object copy is a hack
    # The check for previous occ is not elegant
    overlaps_dict = {}
    for range in virtual_page_ranges:
        base_va = range.va
        off = 0
        for phys_range, phys_range_size in zip(range.phys, range.sizes):
            phys_range_end = phys_range + phys_range_size
            for saved_range in phys_ranges:
                if saved_range[0] > phys_range_end:
                    break
                beg = max(phys_range, saved_range[0]) 
                end = min(phys_range_end, saved_range[1]) 
                va = base_va + off + (beg - phys_range)
                if beg < end and va != saved_range[2]:
                    key = (beg, end)
                    # Make copy and clean-up
                    range_copy = copy.copy(range)
                    range_copy.phys = None
                    range_copy.size = None
                    range_copy.va = va
                    range_copy.page_size = end - beg
                    if key in overlaps_dict:
                        found = False
                        for tmp in overlaps_dict[key]:
                            if tmp.va == va:
                                found = True
                                break
                        if not found:
                            overlaps_dict[key].append(range_copy)
                    else:
                        overlaps_dict[key] = [range_copy]
            off = off + phys_range_size

    # Print the found aliases
    for key in overlaps_dict.keys():
        overlaps = overlaps_dict[key]
        if len(overlaps) > 1:
            print(f"Phys: {hex(key[0])} - {hex(key[1])}")
            overlap_len = key[1] - key[0]
            for overlap in overlaps:
                print(" " * 4 + overlap.to_string(phys_verbose))
            print("")

class PageTableWalkInfo():

    def __init__(self, va):
        self.va = va
        self.faulted = False
        self.stages = []

    def add_register_stage(self, register_name, register_value):
        self.base_register = (register_name, register_value)

    def add_stage(self, stage_str, table_index, entry_value_without_meta, meta_bits):
        self.stages.append((stage_str, table_index, entry_value_without_meta, meta_bits))

    def set_faulted(self):
        self.faulted = True

    def __str__(self):
        s = ""

        s += f"Page table walk for VA = {hex(self.va)}\n"
        s += "-" * 43 + "\n"

        s += f"{self.base_register[0]} = {hex(self.base_register[1])}\n"

        for (stage_index, stage_entry) in enumerate(self.stages):
            stage_str, table_index, entry_value_without_meta, meta_bits = stage_entry
            stage_index = stage_index + 1
            mapping_string = " " * stage_index * 2 + f"{stage_str}[{table_index}] = {hex(entry_value_without_meta)}"
            flags_string = f"Flags 0x{meta_bits:03x}"
            s += mapping_string.ljust(34) + "| " + flags_string + "\n"

        if self.faulted:
            s += "\n!!! Last stage faulted !!!\n"

        return s



from pt.machine import *
from pt.pt import *
from pt.pt_x86_64_parse import *
from pt.pt_aarch64_parse import *
from pt.pt_riscv64_parse import *

import gdb
import os
import subprocess

class QemuGdbMachine(Machine):

    def __init__(self):
        self.pid = QemuGdbMachine.get_qemu_pid()
        self.file = os.open(f"/proc/{self.pid}/mem", os.O_RDONLY)
        self.mem_size = os.fstat(self.file).st_size

    def __del__(self):
        if self.file:
            os.close(self.file)

    def read_physical_memory(self, physical_address, length):
        res = gdb.execute(f"monitor gpa2hva {hex(physical_address)}", to_string = True)

        # It's not possible to pread large sizes, so let's break the request
        # into a few smaller ones.
        max_block_size = 1024 * 1024 * 256
        try:
            hva = int(res.split(" ")[-1], 16)
            data = b""
            for offset in range(0, length, max_block_size):
                length_to_read = min(length - offset, max_block_size)
                block = os.pread(self.file, length_to_read, hva + offset)
                data += block
            return data
        except Exception as e:
            msg = f"Physical address ({hex(physical_address)}, +{hex(length)}) is not accessible. Reason: {e}. gpa2hva result: {res}"
            raise OSError(msg)

    def search_pids_for_file(pids, filename):
        for pid in pids:
            fd_dir = f"/proc/{pid}/fd"

            try:
                for fd in os.listdir(fd_dir):
                    if os.readlink(f"{fd_dir}/{fd}") == filename:
                        return pid
            except FileNotFoundError:
                # Either the process has gone or fds are changing, not our pid
                pass
            except PermissionError:
                # Evade processes owned by other users
                pass

        return None

    @staticmethod
    def get_qemu_pid():
        out = subprocess.check_output(["pgrep", "qemu-system"], encoding="utf8")
        pids = out.strip().split('\n')

        if len(pids) == 1:
            return int(pids[0], 10)

        # We add a chardev file backend (we dont add a fronted, so it doesn't affect
        # the guest). We can then look through proc to find which process has the file
        # open. This approach is agnostic to namespaces (pid, network and mount).
        chardev_id = "gdb-pt-dump" + '-' + ''.join(random.choices(string.ascii_letters, k=16))
        with tempfile.NamedTemporaryFile() as t:
            gdb.execute(f"monitor chardev-add file,id={chardev_id},path={t.name}")
            ret = QemuGdbMachine.search_pids_for_file(pids, t.name)
            gdb.execute(f"monitor chardev-remove {chardev_id}")

        if not ret:
            raise Exception("Could not find qemu pid")

        return int(ret, 10)

    def read_register(self, register_name):
        return int(gdb.parse_and_eval(register_name).cast(gdb.lookup_type("unsigned long")))

class PageTableDumpGdbFrontend(gdb.Command):
    """
    GDB pt-dump: command for inspecting VM page tables.
    Arguments:
        -filter FILTER [FILTER ...]
            Specify filters for the recorded pages.
            x86_64 Supported filters:
                w: is writeable.
                x: is executable
                w|x: is writeable or executable
                ro: read-only
                u: user-space page
                s: supervisor page
                wb: write-back
                uc: uncacheable

             aarch64- and riscv64-supported filters:
                w: is writeable.
                x: is executable
                w|x: is writeable or executable
                ro: read-only
                u: user-space page
                s: supervisor page

        -range START_ADDR END_ADDR
            Will filter-out virtual memory ranges which start at a position in [START_ADDR, END_ADDR]
        -has ADDR
            Will filter-out virtual memory ranges which contain ADDR
        -before ADDR
            Will select virtual memory ranges which start <ADDR
        -after ADDR
            Will select virtual memory ranges which start >=ADDR
        -ss "STRING"
            Searches for the string STRING in the ranges after filtering
        -sb BYTESTRING
            Searches for the byte-string BYTESTRING in the ranges after filtering
        -s8 VALUE
            Searches for the value VALUE in the ranges after filtering
            VALUE should fit in 8 bytes.
        -s4 VALUE
            Searches for the value VALUE in the ranges after filtering
            VALUE should fit in 4 bytes.
        -align ALIGNMENT [OFFSET]
            When searching, it will print out addresses which are aligned to ALIGNMENT.
            If offset is provided, then the check would be performed as (ADDR - OFFSET) % ALIGNMENT.
            It can be useful when searching for content in a particular SLAB.
        -kaslr
            Print KASLR-relevant information like the image offsets and phys map base.
        -kaslr_leaks
            Searchers for values which disclose KASLR offsets.
        -save
            Cache the recorded page table for that address after traversing the hierachy.
            This will yield speed-up when printing the page table again.
        -list
            List the cached page tables.
        -clear
            Clear all saved page tables.
        -info
            Print arch register information.
        -o FILE_NAME
            Store the output from the current command to a file with name FILE_NAME.
            This may be useful when the a lot of data is produced, e.g. full page table.
        -find_alias
            Experimental feature and currently slow. Searches for aliases ranges in virtual memory.
            Ranges are aliased if they point to the the same physical memory. This can be useful if one
            is searching for R/RX memory which is writeable through some other address.
            Another interesting option is to find alias for memory mapped in user space and kernel space.
            TODO: This feature will be reworked for usability and performance in the near future.
        -force_traverse_all
            Forces the traversal of any page table entry (pml4, pdp, ...) even if a duplicate entry has
            already been trarversed. Using this option bypasses an optimization which discards already
            traversed duplicate entries. Expect that using this option would render pt unusable for
            windows VMs.
        -phys_verbose
            Prints the start physical address for the printed virtual ranges. This argument further
            restricts the merging of virtual ranges by requiring that merged ranges need to also be
            physically contiguous. Using this range leads to more verbose output.

    Architecture-specific arguments:
        - X86-32 / X86-64
            `-cr3 HEX_ADDR`
                The GPA of the page table. If not used, the script will use the architectural
                register (e.g. cr3).

        - aarch64
            `-ttbr0_el1 HEX_ADDR`
                The GPA of the TTBR0_EL1 register.
            `-ttbr1_el1 HEX_ADDR`
                The GPA of the TTBR1_EL1 register.

        - riscv64
            `-satp HEX_ADDR`
                The GPA of the SATP register.

    Example usage:
        `pt -save -filter s w|x wb`
            Traverse the current page table and then save it. When returning the result,
            filter the pages to be marked as supervisor, be writeable or executable, and marked as
            write-back.
        `pt -filter w x`
            Traverse the current page table and print out mappings which are both writeable and
            executable.
        `pt -cr3 0x4000`
            Traverse the page table at guest physical address 0x4000. Don't save it.
        `pt -save -kaslr`
            Traverse page tables, save them and print kaslr information.
        `pt -ss "Linux 4."`
            Search for the string Linux.
        `pt -sb da87374107`
            Search for the byte-string da87374107.
        `pt -s8 0xaabbccdd`
            Search for the 8-byte-long value 0xaabbccdd.
        `pt -has 0xffffffffaaf629f7`
            Print information about the mapping which covers the address 0xffffffffaaf629f7.
    """

    def __init__(self):
        super(PageTableDumpGdbFrontend, self).__init__("pt", gdb.COMMAND_USER)
        self.pid = -1
        self.pt = None

    def lazy_init(self):

        # Create machine backend
        machine_backend = QemuGdbMachine()

        # Create arch backend
        arch = gdb.execute("show architecture", to_string = True)
        arch_backend = None
        if "aarch64" in arch:
            arch_backend = PT_Aarch64_Backend(machine_backend)
        elif "x86" in arch or "x64" in arch:
            arch_backend = PT_x86_64_Backend(machine_backend)
        elif "riscv:rv64" in arch:
            arch_backend = PT_RiscV64_Backend(machine_backend)
        else:
            raise Exception(f"Unknown arch. Message: {arch}")

        # Bring-up pt_dump
        self.pt = PageTableDump(machine_backend, arch_backend)

    def invoke(self, arg, from_tty):
        try:
            curr_pid = QemuGdbMachine.get_qemu_pid()
            if curr_pid != self.pid:
                self.lazy_init()
        except Exception as e:
            print("Cannot get qemu-system pid", e)
            return

        argv = gdb.string_to_argv(arg)
        self.pt.handle_command_wrapper(argv)



from pt.pt_common import *
from pt.pt_aarch64_definitions import *
from pt.pt_arch_backend import PTArchBackend
from pt.pt_constants import *
from pt.machine import *

import math

PT_AARCH64_4KB_PAGE  = PT_SIZE_4K
PT_AARCH64_16KB_PAGE = PT_SIZE_16K
PT_AARCH64_64KB_PAGE = PT_SIZE_64K

def is_user_readable(block):
    return block.permissions == 0b11 or block.permissions == 0b01

def is_kernel_readable(block):
    return True

def is_user_writeable(block):
    return block.permissions == 0b01

def is_kernel_writeable(block):
    return block.permissions == 0b01 or block.permissions == 0b00

def is_user_executable(block):
    return (not block.xn)

def is_kernel_executable(block):
    return not block.pxn

class Aarch64_Block():
    def __init__(self, va, phys, size, xn, pxn, permissions):
        self.va = va
        self.page_size = size
        self.xn = xn
        self.pxn = pxn
        self.permissions = permissions
        self.phys = [phys]
        self.sizes = [size]

    def cut_before(self, va):
        print("cut_before is not implemented")

    def cut_after(self, va):
        print("cut_after is not implemented")

    def to_string(self, phys_verbose):
        varying_str = None
        if phys_verbose:
            fmt = f"{{:>{PrintConfig.va_len}}} : {{:>{PrintConfig.page_size_len}}} : {{:>{PrintConfig.phys_len}}} "
            varying_str = fmt.format(hex(self.va), hex(self.page_size), hex(self.phys[0]))
        else:
            fmt = f"{{:>{PrintConfig.va_len}}} : {{:>{PrintConfig.page_size_len}}} "
            varying_str = fmt.format(hex(self.va), hex(self.page_size))

        uspace_writeable = is_user_writeable(self)
        kspace_writeable = is_kernel_writeable(self)
        uspace_readable = is_user_readable(self)
        kspace_readable = is_kernel_readable(self)
        uspace_executable = is_user_executable(self)
        kspace_executable = is_kernel_executable(self)
        delim = bcolors.YELLOW + " " + bcolors.ENDC
        uspace_color = select_color(uspace_writeable, uspace_executable, uspace_readable)
        uspace_str = uspace_color + f" R:{int(uspace_readable)} W:{int(uspace_writeable)} X:{int(uspace_executable)} " + bcolors.ENDC
        kspace_color = select_color(kspace_writeable, kspace_executable, kspace_readable)
        kspace_str = kspace_color + f" R:{int(kspace_readable)} W:{int(kspace_writeable)} X:{int(kspace_executable)}  " + bcolors.ENDC
        s = f"{varying_str}" + delim + uspace_str + delim + kspace_str
        return s

    def read_memory(self, machine):
        memory = b""
        for phys_range_start, phys_range_size in zip(self.phys, self.sizes):
            memory += machine.read_physical_memory(phys_range_start, phys_range_size)
        return memory

    def pwndbg_is_writeable(self):
        return is_user_writeable(self) or is_kernel_writeable(self)

    def pwndbg_is_executable(self):
        return is_user_executable(self) or is_kernel_executable(self)

class Aarch64_Table():
    def __init__(self, pa, va, pxn, xn, permissions):
        self.va = va
        self.pa = pa
        self.permissions = permissions
        self.pxn = pxn
        self.xn = xn

def aarch64_semantically_similar(p1: Aarch64_Block, p2: Aarch64_Block) -> bool:
    return p1.xn == p2.xn and p1.pxn == p2.pxn and p1.permissions == p2.permissions

class PT_Aarch64_Backend(PTArchBackend):
    def __init__(self, machine):
        self.machine = machine
        self.init_registers()

    def init_registers(self):
        self.pt_tcr = PT_TCR(self.machine)

    def print_stats(self):
        print(self.pt_tcr.check())

    def get_arch(self):
        return "aarch64"

    def walk(self, va):
        pt_walk = PageTableWalkInfo(va)

        # Use canonical form to determine which page table to use.
        top_bit_index = 63
        granule_size = None
        as_limit = None

        table_addr = None
        if va & (1 << top_bit_index) == 0:
            # top bit is not set, so this is a userspace address
            granule_size = self.determine_ttbr0_granule_size()
            as_limit = self.determine_ttbr0_address_space_limit()
            table_addr = extract_no_shift(self.get_ttbr0_el1(), 0, 47)
            pt_walk.add_register_stage("TTBR0_EL1", table_addr)
        else:
            granule_size = self.determine_ttbr1_granule_size()
            as_limit = self.determine_ttbr1_address_space_limit()
            table_addr = extract_no_shift(self.get_ttbr1_el1(), 0, 47)
            pt_walk.add_register_stage("TTBR1_EL1", table_addr)

        entry_size = 8
        bits_per_stage = int(math.log2(granule_size / entry_size))
        start_index = int(math.log2(granule_size))
        ranges = reversed([(base, base + bits_per_stage - 1) for base in range(start_index, as_limit, bits_per_stage)])

        for (index, r) in enumerate(ranges):
            page_pa = table_addr & ~0xfff
            entry_index = extract(va, r[0], r[1])
            entry_page_pa = page_pa + entry_index * entry_size
            entry_value = int.from_bytes(self.machine.read_physical_memory(entry_page_pa, entry_size), 'little')
            entry_value_pa = extract_no_shift(entry_value, 0, 47)
            entry_value_pa_no_meta = extract_no_shift(entry_value, 12, 47)
            meta_bits = extract_no_shift(entry_value, 0, 11)

            pt_walk.add_stage(f"Level{index}", entry_index, entry_value_pa_no_meta, meta_bits)

            bit1and2 = extract(entry_value, 0, 1)
            is_valid = bit1and2 != 0
            if not is_valid:
                pt_walk.set_faulted()
                break

            is_block = bit1and2 == 0x1
            if is_block:
                break
            table_addr = entry_value_pa

        return pt_walk

    def get_filter_is_writeable(self, has_superuser_filter, has_user_filter):
        if has_superuser_filter == True and has_user_filter == False:
            return lambda p: is_kernel_writeable(p)
        elif has_superuser_filter == False and has_user_filter == True:
            return lambda p: is_user_writeable(p)
        else:
            return lambda p: is_kernel_writeable(p) or is_user_writeable(p)

    def get_filter_is_not_writeable(self, has_superuser_filter, has_user_filter):
        if has_superuser_filter == True and has_user_filter == False:
            return lambda p: not is_kernel_writeable(p)
        elif has_superuser_filter == False and has_user_filter == True:
            return lambda p: not is_user_writeable(p)
        else:
            return lambda p: not is_kernel_writeable(p) and not is_user_writeable(p)

    def get_filter_is_executable(self, has_superuser_filter, has_user_filter):
        if has_superuser_filter == True and has_user_filter == False:
            return lambda p: is_kernel_executable(p)
        elif has_superuser_filter == False and has_user_filter == True:
            return lambda p: is_user_executable(p)
        else:
            return lambda p: is_user_executable(p) or is_kernel_executable(p)

    def get_filter_is_not_executable(self, has_superuser_filter, has_user_filter):
        if has_superuser_filter == True and has_user_filter == False:
            return lambda p: not is_kernel_executable(p)
        elif has_superuser_filter == False and has_user_filter == True:
            return lambda p: not is_user_executable(p)
        else:
            return lambda p: not is_user_executable(p) and not is_kernel_executable(p)

    def get_filter_is_writeable_or_executable(self, has_superuser_filter, has_user_filter):
        if has_superuser_filter == True and has_user_filter == False:
            return lambda p: is_kernel_writeable(p) or is_kernel_executable(p)
        elif has_superuser_filter == False and has_user_filter == True:
            return lambda p: is_user_writeable(p) or is_user_executable(p)
        else:
            return lambda p: is_kernel_writeable(p) or is_kernel_executable(p) or \
                                is_user_writeable(p) or is_user_executable(p)

    def get_filter_is_user_page(self, has_superuser_filter, has_user_filter):
        return lambda p: is_user_writeable(p) or is_user_readable(p) or is_user_executable(p)

    def get_filter_is_superuser_page(self, has_superuser_filter, has_user_filter):
        return lambda p: is_kernel_writeable(p) or is_kernel_readable(p) or is_kernel_executable(p)

    def get_filter_is_read_only_page(self, has_superuser_filter, has_user_filter):
        l_kernel = lambda p: (not is_kernel_writeable(p) and not is_kernel_executable(p)) and is_kernel_readable(p)
        l_user = lambda p: (not is_user_writeable(p) and not is_user_executable(p)) and is_user_readable(p)
        if has_superuser_filter == True and has_user_filter == False:
            return l_kernel
        elif has_superuser_filter == False and has_user_filter == True:
            return l_user
        else:
            return lambda p: l_user(p) or l_kernel(p)

    def get_filter_architecture_specific(self, filter_name, has_superuser_filter, has_user_filter):
        raise exception(f"Uknown filter {filter_name}")

    def get_ttbr0_el1(self):
        return self.machine.read_register("$TTBR0_EL1")

    def get_ttbr1_el1(self):
        return self.machine.read_register("$TTBR1_EL1")

    def determine_ttbr0_granule_size(self):
        tb0_granule_size = None
        tg0 = self.pt_tcr.TG0
        if tg0 == 0b00:
            tb0_granule_size = PT_AARCH64_4KB_PAGE
        elif tg0 == 0b01:
            tb0_granule_size = PT_AARCH64_64KB_PAGE
        elif tg0 == 0b10:
            tb0_granule_size = PT_AARCH64_16KB_PAGE
        else:
            raise Exception(f"Unknown TG0 value {tg0}")

        return tb0_granule_size

    def determine_ttbr1_granule_size(self):
        tb1_granule_size = None
        tg1 = self.pt_tcr.TG1
        if tg1 == 0b10:
            tb1_granule_size = PT_AARCH64_4KB_PAGE
        elif tg1 == 0b11:
            tb1_granule_size = PT_AARCH64_64KB_PAGE
        elif tg1 == 0b01:
            tb1_granule_size = PT_AARCH64_16KB_PAGE
        else:
            raise Exception(f"Unknown TG1 value {tg1}")

        return tb1_granule_size

    def determine_ttbr0_address_space_limit(self):
        return 64 - self.pt_tcr.T0SZ

    def determine_ttbr1_address_space_limit(self):
        return 64 - self.pt_tcr.T1SZ

    def aarch64_parse_entries(self, tbl, level_range, as_size, granule, is_last_level):
        # lvl starts from one to be in sync with the armv7 docs
        entries = []
        start_bit = int(math.log2(granule))

        try:
            entries = split_range_into_int_values(read_arbitrary_page(self.machine, tbl.pa, granule), 8)
        except Exception:
            pass

        tables = []
        blocks = []
        for i, pa in enumerate(entries):
            is_valid = bool(pa & 0x1)
            if is_valid:
                bit1 = extract(pa, 1, 1)
                bit1and2 = extract(pa, 0, 1)
                contiguous_bit = extract(pa, 52, 52)
                is_block_or_page = (bit1and2 == 1) or is_last_level
                is_table = (not is_block_or_page)
                address_contrib = (i << level_range[0])
                child_va = tbl.va | address_contrib
                if is_table:
                    next_level_address = extract_no_shift(pa, start_bit, 47)
                    permissions = extract(pa, 61, 62)
                    xn = (extract(pa, 60, 60) == 0x1) | tbl.xn
                    pxn = extract(pa, 59, 59) == 0x1 | tbl.pxn
                    tables.append(Aarch64_Table(next_level_address, child_va, pxn, xn, permissions))
                else:
                    xn = (extract(pa, 54, 54) == 0x1) | tbl.xn
                    pxn = (extract(pa, 53, 53) == 0x1) | tbl.pxn
                    phys_addr = extract_no_shift(pa, start_bit, 47)
                    permissions = extract(pa, 6, 7)
                    size = (1 << level_range[0])
                    blocks.append(Aarch64_Block(child_va, phys_addr, size, xn, pxn, permissions))

        return tables, blocks

    def arm_traverse_table(self, pt_addr, as_size, granule_size, leading_bit):
        num_entries_in_page = int(granule_size / 8)
        level_bit_coverage = int(math.log2(num_entries_in_page))
        low_bit_inclusive = int(math.log2(granule_size))
        top_bit_inclusive = as_size - 1

        table_ranges = list(reversed([(low, min(low + level_bit_coverage, top_bit_inclusive)) for low in range(low_bit_inclusive, top_bit_inclusive, level_bit_coverage)]))

        root = Aarch64_Table(pt_addr, 0, 0, 0, 0)
        tables = [root]
        all_blocks = []
        for (level, address_range) in enumerate(table_ranges):
            is_last_level = (level + 1) == len(table_ranges)
            new_tables = []
            for table in tables:
                cur_tables, cur_blocks = self.aarch64_parse_entries(table, address_range, as_size, granule_size, is_last_level)
                new_tables.extend(cur_tables)
                all_blocks.extend(cur_blocks)
            tables = new_tables

        if leading_bit == 1:
            for block in all_blocks:
                block.va = make_canonical(block.va | (1<<as_size), as_size+1)

        return all_blocks

    def parse_tables(self, cache, args):
        requires_physical_contiguity = args.phys_verbose
        tb0 = args.ttbr0_el1
        tb1 = args.ttbr1_el1

        if tb0:
            tb0 = int(tb0[0], 16)
        if tb1:
            tb1 = int(tb1[0], 16)

        # If neither is provided, just then query the actual translation table registers.
        if not tb0 and not tb1:
            tb0 = self.get_ttbr0_el1()
            tb1 = self.get_ttbr1_el1()

        all_blocks_0 = []
        all_blocks_1 = []
        if not args.ttbr0_el1:
            # Try to get the blocks from TTBR0 only if user has not specifically requested interpreting
            # the provided TTBR1.
            tb0 = extract_no_shift(tb0, 10, self.determine_ttbr0_address_space_limit() - 1)
            if tb0 in cache:
                all_blocks_0 = cache[tb0]
            else:
                tb0_granule_size = self.determine_ttbr0_granule_size()
                tb0_sz = self.determine_ttbr0_address_space_limit()
                all_blocks_0 = self.arm_traverse_table(tb0, tb0_sz, tb0_granule_size, 0)
                all_blocks_0 = optimize([], [], all_blocks_0, aarch64_semantically_similar, requires_physical_contiguity)

        if not args.ttbr1_el1:
            tb1 = extract_no_shift(tb1, 10, self.determine_ttbr1_address_space_limit() - 1)
            if tb1 in cache:
                all_blocks_1 = cache[tb1]
            else:
                tb1_granule_size = self.determine_ttbr1_granule_size()
                tb1_sz = self.determine_ttbr1_address_space_limit()
                all_blocks_1 = self.arm_traverse_table(tb1, tb1_sz, tb1_granule_size, 1)
                all_blocks_1 = optimize([], [], all_blocks_1, aarch64_semantically_similar, requires_physical_contiguity)

        # TODO: Consider the top-byte ignore rules
        # TODO: Consider EPDs

        if args.save:
            cache[tb0] = all_blocks_0
            cache[tb1] = all_blocks_1

        all_blocks = all_blocks_0 + all_blocks_1

        return all_blocks

    def print_kaslr_information(self, table, should_print = True, phys_verbose = False):
        potential_base_filter = lambda p: is_kernel_executable(p)
        tmp = list(filter(potential_base_filter, table))
        found_page = None
        kaslr_addresses = []
        stext_phys_base_addr = 0x40210000
        text_phys_base_addr = 0x40200000
        for page in tmp:
            if page.phys[0] == stext_phys_base_addr or page.phys[0] == text_phys_base_addr:
                found_page = page
                break
        if found_page:
            kaslr_addresses.append(found_page.va)
            if should_print:
                print("Found virtual image base:")
                print("\tVirt: " + found_page.to_string(phys_verbose))
                print("\tPhys: " + hex(found_page.phys[0]))
        else:
            if should_print:
                print("Failed to determine kaslr offsets")
        return kaslr_addresses

    def print_table(self, table, phys_verbose):
        varying_str = None
        if phys_verbose:
            fmt = f"{{:>{PrintConfig.va_len}}} : {{:>{PrintConfig.page_size_len}}} : {{:>{PrintConfig.phys_len}}} |"
            varying_str = fmt.format("Address", "Length", "Phys")
        else:
            fmt = f"{{:>{PrintConfig.va_len}}} : {{:>{PrintConfig.page_size_len}}} |"
            varying_str = fmt.format("Address", "Length")
        print(bcolors.BLUE + varying_str + " User space " + " | Kernel space " + bcolors.ENDC)
        for block in table:
            print(block.to_string(phys_verbose))




from pt.pt_register import *

class PT_CR0(PT_Register):
    def __init__(self, machine):
        super(PT_CR0, self).__init__(machine, "cr0", "Control Register 0")
        self.add_range("PE (Protected Mode Enable)", 0, 0, PT_Decipher_Meaning_Match({0: "Protected mode", 1: "Real mode"}))
        self.add_range("MP (Monitor co-processor)", 1, 1, PT_Decipher_Meaning_Passthrough)
        self.add_range("EM (Emulation)", 2, 2, PT_Decipher_Meaning_Match({1: "No x87 FPU present", 0: "x87 FPU present"}))
        self.add_range("TS (Task switched)", 3, 3, PT_Decipher_Meaning_Passthrough)
        self.add_range("ET (Extension type)", 4, 4, PT_Decipher_Meaning_Passthrough)
        self.add_range("NE (Numeric error)", 5, 5, PT_Decipher_Meaning_Passthrough)
        self.add_range("WP (Write protect)", 16, 16, PT_Decipher_Meaning_Passthrough)
        self.add_range("AM (Alignment mask)", 18, 18, PT_Decipher_Meaning_Passthrough)
        self.add_range("NW (Not write-through)", 29, 29, PT_Decipher_Meaning_Passthrough)
        self.add_range("CD (Cache disable)", 30, 30, PT_Decipher_Meaning_Passthrough)
        self.add_range("PG (Paging)", 31, 31, PT_Decipher_Meaning_Passthrough)

class PT_CR4(PT_Register):
    def __init__(self, machine):
        super(PT_CR4, self).__init__(machine, "cr4", "Control Register 4")
        self.add_range("VME (Virtual 8086 Mode Extensions)", 0, 0, PT_Decipher_Meaning_Passthrough)
        self.add_range("PVI (Protected-model virtual interrupts)", 1, 1, PT_Decipher_Meaning_Passthrough)
        self.add_range("TSD (Time Stamp Disable)", 2, 2, PT_Decipher_Meaning_Passthrough)
        self.add_range("DE (Debugging Extensions)", 3, 3, PT_Decipher_Meaning_Passthrough)
        self.add_range("PSE (Page Size Extension)", 4, 4, PT_Decipher_Meaning_Passthrough)
        self.add_range("PAE (Physical Address Extension)", 5, 5, PT_Decipher_Meaning_Passthrough)
        self.add_range("MCE (Machine Check Exception)", 6, 6, PT_Decipher_Meaning_Passthrough)
        self.add_range("PGE (Page Global Enabled)", 7, 7, PT_Decipher_Meaning_Passthrough)
        self.add_range("PCE (Performance Monitor Counter Enable)", 8, 8, PT_Decipher_Meaning_Passthrough)
        self.add_range("OSFXSR", 9, 9, PT_Decipher_Meaning_Passthrough)
        self.add_range("OSXMMEXCPT", 10, 10, PT_Decipher_Meaning_Passthrough)
        self.add_range("UMIP (User mode instruction prevention)", 11, 11, PT_Decipher_Meaning_Passthrough)
        self.add_range("LA57", 12, 12, PT_Decipher_Meaning_Passthrough)
        self.add_range("VMXE", 13, 13, PT_Decipher_Meaning_Passthrough)
        self.add_range("SMXE", 14, 14, PT_Decipher_Meaning_Passthrough)
        self.add_range("FSGSBASE", 16, 16, PT_Decipher_Meaning_Passthrough)
        self.add_range("PCIDE", 17, 17, PT_Decipher_Meaning_Passthrough)
        self.add_range("OSXSAVE", 18, 18, PT_Decipher_Meaning_Passthrough)
        self.add_range("SMEP", 20, 20, PT_Decipher_Meaning_Passthrough)
        self.add_range("SMAP", 21, 21, PT_Decipher_Meaning_Passthrough)
        self.add_range("PKE", 22, 22, PT_Decipher_Meaning_Passthrough)




from collections import namedtuple
from pt.pt_common import *

PT_Register_Range = namedtuple('PT_Register_Range', ['name', 'low', 'high', 'func'])

class PT_Register_State:
    def __init__(self, short_name, name, kv):
        self.short_name = short_name
        self.name = name
        self.kv = kv

    def __str__(self):
        s = ""
        total = 148
        s += bcolors.BLUE + f"{self.short_name} ({self.name}):".ljust(total) +  bcolors.ENDC  + "\n"
        delim = "|"
        for key in self.kv:
            value, low, high, res  = self.kv[key]
            s += f"    {key}".ljust(10) + " (" + f"{low}".rjust(2) + ":" + f"{high}".rjust(2) + ") = " + hex(res).rjust(4) + " " + delim + f" {value} ".ljust(128) + "\n"
        s += "-" * total + "\n"
        return s

    def get_value(self, key):
        return self.kv[key][3]

class PT_Decipher_Meaning_Match:
    def __init__(self, kv):
        self.kv = kv

    def __call__(self, key):
        return self.kv[key]

PT_Decipher_Meaning_Passthrough = lambda x: x

class PT_Register:
    def __init__(self, machine, register, name):
        self.machine = machine
        self.register = register
        self.name = name
        self.ranges_dict = {}

    def add_range(self, name, low, high, decipher_meaning):
        self.ranges_dict[name] = PT_Register_Range(name = name, low = low, high = high, func = decipher_meaning)

    def check(self):
        reg_value = self.machine.read_register(f"${self.register}")
        kv = dict()
        for key in self.ranges_dict:
            r = self.ranges_dict[key]
            res = extract(reg_value, r.low, r.high)
            kv[r.name] = (r.func(res), r.low, r.high, res)
        return PT_Register_State(self.register, self.name, kv)

    def __getattr__(self, attr):
        return self.check().get_value(str(attr))




from abc import ABC
from abc import abstractmethod

class Machine(ABC):

    def __init__(self):
        pass

    @abstractmethod
    def read_register(self, register_name):
        raise Exception("Unimplemented")

    @abstractmethod
    def read_physical_memory(self, physical_address, length):
        raise Exception("Unimplemented")




from pt.pt_common import *

class PML4_Entry():
    def __init__(self, value, index):
        self.present = is_present(value)
        self.writeable = is_writeable(value)
        self.supervisor = is_supervisor(value)
        self.writeback = is_writeback(value)
        self.cacheable = is_cacheable(value)
        self.accessed = is_accessed(value)
        self.available = is_available(value)
        self.nx = is_nx(value)
        self.pdp = get_pdp_base(value)
        self.raw = value
        self.virt_part = (index << 39)

    def __str__(self):
        res = (f"{hex(self.pdp)}: "
               f"P:{int(self.present)} "
               f"W:{int(self.writeable)} "
               f"S:{int(self.supervisor)} "
               f"WB:{int(self.writeback)} "
               f"UC:{int(not self.cacheable)} "
               f"A:{int(self.accessed)} "
               f"AVL:{int(self.available)} "
               f"NX:{int(self.nx)}")
        return res

class PDP_Entry():
    def __init__(self, value, parent_va, index):
        self.present = is_present(value)
        self.writeable = is_writeable(value)
        self.supervisor = is_supervisor(value)
        self.writeback = is_writeback(value)
        self.cacheable = is_cacheable(value)
        self.accessed = is_accessed(value)
        self.virt_part = (index << 30) | parent_va
        self.large_page = is_large_page(value) # This means it's a leaf
        if self.large_page:
            self.dirty = is_dirty(value)
            self.glob = True
            self.pd = extract_no_shift(value, 30, 51)
        else:
            self.pd = get_pdp_base(value)
        self.nx = is_nx(value)

    def __str__(self):
        res = (f"{hex(self.pd)}: "
               f"P:{int(self.present)} "
               f"W:{int(self.writeable)} "
               f"S:{int(self.supervisor)} "
               f"WB:{int(self.writeback)} "
               f"UC:{int(not self.cacheable)} "
               f"A:{int(self.accessed)} ")
        if self.large_page:
                res += (f"D:{int(self.dirty)} "
                        f"G:{int(self.glob)} "
                        f"NX:{int(self.nx)}")
        return res

class PD_Entry():
    def __init__(self, value, parent_va, index, pde_shift):
        self.present = is_present(value)
        self.writeable = is_writeable(value)
        self.supervisor = is_supervisor(value)
        self.writeback = is_writeback(value)
        self.cacheable = is_cacheable(value)
        self.accessed = is_accessed(value)
        self.virt_part = (index << pde_shift) | parent_va
        self.big_page = is_big_page(value) # This means it's a leaf
        if self.big_page:
            self.dirty = is_dirty(value)
            self.glob = True
            self.pat = is_pat(value)
            # TODO
            self.pt = extract_no_shift(value, 20, 51)
        else:
            self.pt = get_pdp_base(value)
        self.page_size = 1 << pde_shift
        self.nx = is_nx(value)

    def __str__(self):
        res = (f"{hex(self.pt)}: "
               f"P:{int(self.present)} "
               f"W:{int(self.writeable)} "
               f"S:{int(self.supervisor)} "
               f"WB:{int(self.writeback)} "
               f"UC:{int(not self.cacheable)} "
               f"A:{int(self.accessed)} ")
        if self.big_page:
                res += (f"D:{int(self.dirty)} "
                        f"G:{int(self.glob)} "
                        f"NX:{int(self.nx)}")
        return res

class PT_Entry():
    def __init__(self, value, parent_va, index):
        self.present = is_present(value)
        self.writeable = is_writeable(value)
        self.supervisor = is_supervisor(value)
        self.writeback = is_writeback(value)
        self.cacheable = is_cacheable(value)
        self.accessed = is_accessed(value)
        self.dirty = is_dirty(value)
        self.glob = True
        self.pat = is_pat(value)
        self.pt = extract_no_shift(value, 12, 51)
        self.virt = (index << 12) | parent_va
        self.nx = is_nx(value)

    def __str__(self):
        res = (f"{hex(self.pt)}: "
               f"P:{int(self.present)} "
               f"W:{int(self.writeable)} "
               f"S:{int(self.supervisor)} "
               f"WB:{int(self.writeback)} "
               f"UC:{int(not self.cacheable)} "
               f"A:{int(self.accessed)} "
               f"D:{int(self.dirty)} "
               f"PAT:{int(self.pat)} "
               f"G:{int(self.glob)} "
               f"NX:{int(self.nx)}")
        return res

def create_page_from_pte(pte: PT_Entry) -> Page:
    page = Page()
    page.va = make_canonical(pte.virt)
    page.page_size = 4096
    page.w = pte.writeable
    page.x = not pte.nx
    page.s = pte.supervisor
    page.uc = not pte.cacheable
    page.wb = pte.writeback
    page.phys = [pte.pt]
    page.sizes = [page.page_size]
    return page

def create_page_from_pde(pde: PD_Entry) -> Page:
    page = Page()
    page.va = make_canonical(pde.virt_part)
    page.page_size = pde.page_size
    page.w = pde.writeable
    page.x = not pde.nx
    page.s = pde.supervisor
    page.uc = not pde.cacheable
    page.wb = pde.writeback
    page.phys = [pde.pt]
    page.sizes = [page.page_size]
    return page

def create_page_from_pdpe(pdpe: PDP_Entry) -> Page:
    page = Page()
    page.va = make_canonical(pdpe.virt_part)
    page.page_size = 1024 * 1024 * 1024
    page.w = pdpe.writeable
    page.x = not pdpe.nx
    page.s = pdpe.supervisor
    page.uc = not pdpe.cacheable
    page.wb = pdpe.writeback
    page.phys = [pdpe.pd]
    page.sizes = [page.page_size]
    return page

def is_present(addr):
    return (addr & 0x1) != 0

def is_writeable(addr):
    return (addr & 0x2) != 0

def is_supervisor(addr):
    return (addr & 0x4) == 0

def is_writeback(addr):
    return (addr & 0x8) == 0

def is_cacheable(addr):
    return (addr & 0x10) == 0

def is_accessed(addr):
    return (addr & 0x10) == 1

def is_dirty(addr):
    return ((addr >> 6) & 0x1) == 0

def is_available(addr):
    return ((addr >> 9) & 0x3) != 0

def is_nx(addr):
    return (addr & (1<<63)) != 0

def get_pdp_base(addr):
    return extract_no_shift(addr, 12, 51)

# One gigabyte-large page.
def is_large_page(addr):
    return (addr >> 0x7) & 0x1

# Either two-mb- or four-mb-large page.
def is_big_page(addr):
    return (addr >> 7) & 0x1

def is_pat(addr):
    return (addr >> 12) & 0x1

def is_global(addr):
    return (addr >> 0x8) & 0x1

def is_pat(addr):
    return (addr >> 12) & 0x1

def rwxs_semantically_similar(p1: Page, p2: Page) -> bool:
    return p1.w == p2.w and p1.x == p2.x and p1.s == p2.s and p1.wb == p2.wb and p1.uc == p2.uc

